/*
 * module:
 *	ed_tio.c	version 1.1
 *
 * purpose:
 *	this is a machine-dependent module to handle keyboard input and 
 *	console output for the editor
 *
 * external entry points:
 *	tio_init	initialize the output stuff
 *	tio_clup	clean up the output stuff
 *	readkbchr	read a character from the keyboard
 *	outline		output a line to the console
 *	outstr		output a string to the console
 *	outnum		output a decimal integer
 *	outchr		output a character to the console
 *	bell		ring a bell to warn of an error
 *	bug		log an error message
 *
 * internal entry points:
 *	cons_putc	tty-like output routine for PC screen
 *	kbread		input routine for PC keyboard
 *
 * note:
 *	dending on how much you worry about memory, you can compile the
 *	support for the PC console in or out with the pre-processor symbol
 *	CONSOLE.  If this is not defined, the code to support the console
 *	will be left out of the editor.
 */

#include "ed_rtns.h"
#include "ed_state.h"

#ifdef	UNIX
#include <termio.h>
#include <sys/ioctl.h>
struct	termio oldmodes, newmodes;
#endif	UNIX

#ifdef	DOS
#include <dos.h>
#endif	DOS

#ifdef	CONSOLE
int	scrnatr;	/* display attribute on current screen	*/
int	scrnpage;	/* display page for current screen	*/
#endif	CONSOLE
/*
 * routine:
 *	tio_init
 *
 * purpose:
 *	to perform any necessary terminal/screen initialization
 *
 * returns:
 *	number of errors
 */
tio_init()
 {	int errs = 0;

#ifdef	CONSOLE
	union REGS regs;

	if (c_console)
	{	/* figure out where the cursor is and what attribute we use */
		regs.h.ah = 15;	/* BIOS get video state */
		int86( 0x10, &regs, &regs );
		scrnpage = regs.h.bh;

		regs.h.ah = 3;	/* BIOS get cursor position */
		int86( 0x10, &regs, &regs );
		scrnrow = regs.h.dh;
		scrncol = regs.h.dl;
		
		regs.h.ah = 8;	/* BIOS read character/attribute */
		regs.h.bh = scrnpage;
		int86( 0x10, &regs, &regs );
		scrnatr = regs.h.ah;
	}
#endif	CONSOLE
#ifdef	UNIX
	/* get the old modes, turn off echoing and cooking */
	if (ioctl( 0, TCGETA, &oldmodes ) == 0)
	{	newmodes = oldmodes;
		newmodes.c_lflag &= ~(ECHO+ISIG+ICANON);
		newmodes.c_cc[4] = 1;
		newmodes.c_cc[5] = 0;
		if (ioctl( 0, TCSETA, &newmodes ))
			errs++;
	} else
		errs++;
#endif	UNIX

	return( errs );
 }

 /*
  * routine:
  *	tio_clup
  *
  * purpose:
  *	to perform any necessary terminal/screen cleanup
  */
void tio_clup()
{	
	/* restore attribute to normal */
	set_atr( '0' );
	curatr = 0;

	/* clean up the screen */
	term_init( 1 );

#ifdef	UNIX
	/* restore terminal modes upon entry */
	ioctl( 0, TCSETA, &oldmodes );
#endif	UNIX
}


/*
 * routine:
 *	readkbchar
 *
 * purpose:
 *	to read a character from the keyboard
 *
 * parms:
 *	none
 *
 * returns:
 *	> 0 byte read
 *	-1  hangup has taken place
 *
 * algorithm:
 *	read and buffer up a few characters, passing them back to the
 *	caller one at a time.
 */
int readkbchar()
{
	static int count = 0;
	static int pos = 0;
	static char inbuf[10];

	if (pos >= count)
	{	
#ifdef DOS
		count = kbread( inbuf, sizeof inbuf );
#else	!DOS
		count = read( 0, inbuf, sizeof inbuf );
#endif	!DOS

		/* check for input error or end of file */
		if (count <= 0)
			return( -1 );

		/* reste the input buffer pointer	*/
		pos = 0;
	}

	return( inbuf[ pos++ ]  & 0x7f );
}

/*
 * routine:
 *	outline
 *
 * purpose:
 *	to send a line to the terminal
 *
 * parms:
 *	pointer to line to output
 *	number of bytes to write
 *
 * note:
 *	it is assumed (for column counting purposes) that this routine
 *	is only used to output text characters, and not escape sequences.
 */
void outline( buf, len )
 char *buf;
 int len;
{	register char *s;

#ifdef	CONSOLE
	/* for the console - we must process each and every character */
	if (c_console)
	{	/* interpret each character for ourselves */
		for( s = buf; s < &buf[len]; s++ )
			outchr( *s );
	} else
#endif	CONSOLE
		(void) write( 1, buf, len );

	/* in either case, we must figure out what we've done */
	for( s = buf; s < &buf[len]; s++ )
	{	if (*s == '\r')
			scrncol = 0;
		else if (*s == '\n')
			scrnrow++;
		else if (*s >= ' ')
		{	if (++scrncol >= maxcols)
			{	scrncol -= maxcols;
			scrnrow++;
			}	
		}
	}
}

/*
 * routine:
 *	outchr
 *
 * purpose:
 *	to send a character to the terminal
 *
 * parm:
 *	character to output
 *
 */
void outchr( c )
 char c;
{	
#ifdef	CONSOLE
	void cons_putc();

	if (c_console)
		cons_putc( c );
	else
#endif	CONSOLE
		(void) write( 1, &c, 1 );
}

/*
 * routine:
 *	outstr
 *
 * purpose:
 *	to send a null-terminated string to the terminal
 *
 * parm:
 *	pointer to string to send
 *
 * returns:
 *	number of bytes printed
 *
 * note:
 *	this routine assumes that outline will handle the cursor position
 *	implications of this output.
 */
int outstr( s )
 register char *s;
{	register int cnt = strlen(s);

	outline( s, cnt );
	return( cnt );
}

/*
 * routine:
 *	outnum
 *
 * purpose:
 *	to print out a decimal integer
 *
 * parm:
 *	number to print
 *
 * returns:
 *	number of characters printed
 */
int outnum( n )
 register int n;
{	int chars = 0;
	int divisor;

	/* check for zero */
	if (n == 0)
	{	outchr( '0' );
		return( 1 );
	}

	/* check for negative numbers */
	if (n < 0)
	{	outchr( '-' );
		chars++;
		n = -n;
	}

	/* limit the size of the numbers */
	if (n > 10000)
		n %= 10000;

	/* find the order of the first digit */
	for( divisor = 10000; divisor > n; divisor /= 10 );

	do {	outchr( '0' + (n / divisor) );
		chars++;
		n %= divisor;
		divisor /= 10;
	} while( divisor >= 1 );

	return( chars );
}

/*
 * routine:
 *	bell
 *
 * purpose:
 *	to ring a bell at the terminal and warn of an error
 */
void bell()
{
#ifdef	CONSOLE
	if (c_console)
	{	union REGS regs;

		regs.h.ah = 14;	/* BIOS tty-write to screen */
		regs.h.al = 7;	/* bios interprets ^G as bell */
		int86( 0x10, &regs, &regs );
	} else
#endif CONSOLE
		outchr( 007 );
}

/*
 * routine:
 *	bug
 *
 * purpose:
 *	log a message complaining about an internal error
 *
 * parm:
 *	string describing the problem
 */
void bug( s )
 char *s;
{
	bell();
	(void) outstr( "\nERROR: " );
	(void) outstr( s );
	(void) outstr( "\r\n" );
}

#ifdef	DOS
#ifdef	CONSOLE
/*
 * routine:
 *	cons_putc
 *
 * purpose:
 *	to handle tty-like output to the console
 *
 * parm:
 *	character to be written
 *
 * notes:
 *	we could do all of this using int 10 subcode 14 if it only
 *	took attributes.  This is routine simulates cariage return
 *	and line feed and passes everything else to the screen as a
 *	graphic - using the current attribute in scrnatr.
 */
void cons_putc( c )
{	union REGS regs;
	register int row, col;

	/* if we aren't properly initialized yet, just do a write */
	if (scrnatr == 0)
	{	(void) write( 1, &c, 1 );
		return;
	}

	/* find out the current cursor position */
	regs.h.ah = 3;	/* read cursor position */
	regs.h.bh = scrnpage;
	int86( 0x10, &regs, &regs );
	row = regs.h.dh;
	col = regs.h.dl;

	/* the only possible characters are cr, lf, and graphics */
	if (c == '\r')
		col = 0;
	else if (c == '\n')
		row++;
	else
	{	regs.h.ah = 9;	/* write character and attribute */
		regs.h.al = c;
		regs.h.bh = scrnpage;
		regs.h.bl = scrnatr;
		regs.x.cx = 1;
		int86( 0x10, &regs, &regs );

		if (++col >= maxcols)
		{	col = 0;
			row++;
		}
	}

	/* see if we need to scroll another row onto the screen */
	if (row >= maxrows)
	{	regs.h.ah = 6;		/* scroll up */
		regs.h.al = 1 + row - maxrows;
		regs.h.ch = 0;		/* from top left */
		regs.h.cl = 0;
		regs.h.dh = maxrows-1;	/* to bot right */
		regs.h.dl = maxcols-1;
		regs.h.bh = scrnatr;	/* using current attribute */
		int86( 0x10, &regs, &regs );
		row = maxrows - 1;
	}

	/* then put the cursor where it belongs */
	regs.h.ah = 2;		/* pos cursor */
	regs.h.dh = row;
	regs.h.dl = col;
	regs.h.bh = scrnpage;
	int86( 0x10, &regs, &regs );
}

/*
 * array:
 *	scanmap
 *
 * purpose:
 *	to map scan codes into ascii sequences (following esc [)
 *
 * note:
 *	most of the cursor-pad keys return reasonable ascii codes.
 *	the function keys return their numbers, and the remaining
 *	keys return values that made sense to me at the time.
 */
char scanmap[] =
{	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,	/*  00 - 09  */
 	  0,  0,  0,  0,  0,'Z',  0,  0,  0,  0,	/*  10 - 19  */
 	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,	/*  20 - 29  */
 	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,	/*  30 - 39  */
 	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,	/*  40 - 49  */
 	  0,  0,  0,  0,  0,  0,  0,  0,  0,'1',	/*  50 - 59  */
 	'2','3','4','5','6','7','8','9','0',  0,	/*  60 - 69  */
 	  0,'H','A','U',  0,'D',  0,'C',  0,'K',	/*  70 - 79  */
	'B','V','@',127,  0,'$',  0,  0,  0,  0,	/*  80 - 89  */
 	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,	/*  90 - 99  */
 	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,	/* 100 - 109 */
 	  0,  0,  0,  0,  0,'d','c','k',  0,'G',	/* 110 - 119 */
 	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,	/* 120 - 129 */
 	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0		/* 130 - 139 */
};
#endif	CONSOLE

/*
 * routine:
 *	kbread
 *
 * purpose:
 *	to handle dos console input
 *
 * parms:
 *	address and size of input buffer
 *
 * returns:
 *	number of characters read
 */
int kbread( buf, len )
 char *buf;
 int len;
{	register char c;

	/* loop until we get something reasonable */
	for(;;)
	{	/* read a character from the keyboard */
		c = getch();

		/* normal ascii characters are easy	*/
		if (c > 0)
		{	buf[0] = c;
			return( 1 );
		}

#ifdef CONSOLE
		/* null byte means scan-code follows in next byte */
		c = getch();
		c = scanmap[ c ];	/* see if we know it */

		if (c == 0)		/* if not, try again */
			continue;

		if (c == 0x7f)		/* DEL is ascii */
		{	buf[0] = c;
			return( 1 );
		}

		/* anything else comes back as a 3-byte esc sequence */
		buf[0] = '\033';
		buf[1] = '[';
		buf[2] = c;
		return( 3 );
#endif	CONSOLE
	}
}
#endif	DOS
